import tempfile


def WrittenNamedTemporaryFile(
    data,
    mode='w+b',
    buffering=-1,
    encoding=None,
    newline=None,
    suffix=None,
    prefix=None,
    dir=None,
    delete=True,
    seek=None,
        close=False):
    """Create NamedTeporaryFile and write data on it.

    It works the same as NamedTemporaryFile except that it also writes
    data on temp file, with possibility of changing stream position at
    the start and/or closing temp file (when for example it is not
    deleted and will be used later).

    Write mode must be used here.
    """
    f = tempfile.NamedTemporaryFile(
        mode=mode,
        buffering=buffering,
        encoding=encoding,
        newline=newline,
        dir=dir,
        prefix=prefix,
        suffix=suffix,
        delete=delete)
    f.write(data)
    if seek is not None:
        f.seek(seek)
    if close:
        f.close()
    return f



"""Logging helpers to capture and setup logging."""
import io
import contextlib
from typing import Optional
import logging


LOG_FORMAT = '%(asctime)s %(levelname)s %(filename)s: %(message)s'


def capture_output(
    target: object,
    args: Optional[tuple] = None,
        kwargs: Optional[dict] = None) -> str:
    """Redirect stdout and stderr into string buffer and capture it.

    target object is executed with optional args, kwargs and all stdout/
    stderr that is captured, returned in string form.

    Args:
        target: object to execute, usually a function.
        args: target positional arguments (default: {None})
        kwargs: target keyword arguments (default: {None})
    """
    if not args:
        args = ()
    if not kwargs:
        kwargs = {}

    with io.StringIO() as sio:
        with contextlib.redirect_stdout(sio), contextlib.redirect_stderr(sio):
            target(*args, **kwargs)
            output = sio.getvalue()
            return output


def setup_logger(
        log_level: str = logging.WARNING, fmt: str = LOG_FORMAT) -> None:
    """Set up basic logging configuration."""
    logging.basicConfig(
        format=LOG_FORMAT,
        level=log_level)



"""Methods to parse string patterns."""
from operator import attrgetter


def _join_parent_attrs(
        obj, parent_path, val_path, sep=' / ', _reversed=True):
    # TODO: do we need to handle Falsy values in some special way?
    def get_next_val(parent):
        parent_f = attrgetter(parent_path)
        val_f = attrgetter(val_path)
        while parent:
            # Convert to string is needed when non string value is
            # present.
            yield str(val_f(parent))
            parent = parent_f(parent)
    vals_lst = [val for val in get_next_val(obj)]
    if _reversed:
        vals_lst = reversed(vals_lst)
    return ('%s' % sep).join(vals_lst)



"""Helpers for date manipulation."""
import datetime
from dateutil.relativedelta import relativedelta
from typing import Optional

DATE_FMT = '%Y-%m-%d'
DATETIME_FMT = '%Y-%m-%d %H:%M:%S'


def _modify_date(
    dt: datetime.date,
    modify_fun: callable,
    args: Optional[tuple] = None,
        kwargs: Optional[dict] = None) -> datetime.date:
    if not args:
        args = []
    if not kwargs:
        kwargs = {}
    return modify_fun(dt, *args, **kwargs)


def to_date(
    dt: str,
        fmt: Optional[str] = DATE_FMT) -> datetime.date:
    """Convert date string of custom format to datetime.date object."""
    return datetime.datetime.strptime(dt, fmt).date()


def to_string(
    dt: datetime.date,
        fmt: Optional[str] = DATE_FMT) -> str:
    """Convert datetime.date object to custom format date string."""
    return dt.strftime(fmt)


def get_first_day_date(
    dt: datetime.date,
        months: int = 0) -> datetime.date:
    """Return date by replacing day to be first day of the month.

    Args:
        dt (datetime.date): datetime.date object to modify.
        months (int): x month
            (0 - this month, 1 - next month, -1 - last month, ...)

    Returns:
        datetime.date: modified date as datetime.date object with day
            being first of the month.

    """
    return _modify_date(
        dt,
        lambda dt: dt + relativedelta(day=1, months=months))


def get_last_day_date(
    dt: datetime.date,
        months: int = 0) -> datetime.date:
    """Return date by replacing day to be last day of the month.

    Args:
        dt (datetime.date): datetime.date object to modify.
        months (int): x month
            (0 - this month, 1 - next month, -1 - last month, ...)

    Returns:
        datetime.date: modified date as datetime.date object with day
            being last of the month.

    """
    return _modify_date(
        dt,
        lambda dt: dt + relativedelta(day=1, days=-1, months=(months + 1)))


def get_first_day_date_str(
    dt: str,
    fmt: Optional[str] = DATE_FMT,
    new_fmt: Optional[str] = False,
        months: int = 0) -> str:
    """Return date by replacing day to be first day of the month.

    Args:
        dt (str): date as string to modify.
        fmt (str): format to convert from (default: {DATE_FMT})
        new_fmt (str): format to convert to. If not set, will use
            fmt variable format. (default: {False})
        months (int): x month
            (0 - this month, 1 - next month, -1 - last month, ...)

    Returns:
        str: modified date with day being first of the month.

    """
    if not new_fmt:
        new_fmt = fmt
    date_fmt = to_date(dt, fmt=fmt)
    modified_dt = get_first_day_date(date_fmt, months=months)
    return to_string(modified_dt, fmt=new_fmt)


def get_last_day_date_str(
    dt: datetime.date,
    fmt: Optional[str] = DATE_FMT,
    new_fmt: Optional[str] = False,
        months: int = 0) -> str:
    """Return date by replacing day to be last day of the month.

    Args:
        dt (str): date as string to modify.
        fmt (str): format to convert from (default: {DATE_FMT})
        new_fmt (str): format to convert to. If not set, will use
            fmt variable format. (default: {False})
        months (int): x month
            (0 - this month, 1 - next month, -1 - last month, ...)

    Returns:
        str: modified date with day being last of the month.

    """
    if not new_fmt:
        new_fmt = fmt
    date_fmt = to_date(dt, fmt=fmt)
    modified_dt = get_last_day_date(date_fmt, months=months)
    return to_string(modified_dt, fmt=new_fmt)



"""Helper functions for paths manipulation."""
import os
from contextlib import contextmanager
import inspect


def get_cfp(real: bool = False, fdir: bool = False) -> str:
    """Return caller's current file path.

    Args:
        real: if True, return full path, otherwise relative path
            (default: {False})
        fdir: file's directory path will be returned instead
            (default: {False}).
    """
    frame = inspect.stack()[1]
    p = frame[0].f_code.co_filename
    if real:
        p = os.path.realpath(p)
    if fdir:
        p = os.path.dirname(p)
    return p


@contextmanager
def chdir_tmp(path):
    """Change current working directory temporarily."""
    oldpwd = os.getcwd()
    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(oldpwd)



"""Module to check or manipulate various types."""
import collections
import six


def is_iterable(obj):
    """Return True if object is iterable, but is not a string."""
    return (isinstance(obj, collections.Iterable) and not
            isinstance(obj, six.string_types))



"""Module providing various helper decorators."""
import time
import logging
import inspect
from typing import Optional


class _TimeIt(object):
    """Decorator  class to track function execution time.

    This decorator should probably be used as convenience. If you need
    to benchmark your code, you should probably use standard module
    timeit.
    """

    _time_map = {
        'ms': ('millisecond(s)', 1000),
        's': ('second(s)', 1),
        'm': ('minute(s)', 1/60),
        'h': ('hour(s)', 1/3600),
    }

    def __init__(
        self,
        log_level: Optional[str] = None,
        msg_fmt: str = "'{_f_name}' took {_dur:.2f} {_uom_name}",
            uom: str = 's') -> None:
        """Initialize specified arguments for _TimeIt object.

        Args:
            log_level: logging level to use (default: {None})
            msg_fmt: message format. Possible keys for message format:
                - main:
                  - '_f_name': function name.
                  - '_dur': time it took to execute function (in format
                    specified in time_fmt).
                  - '_uom_name': unit of measure name (second(s),
                    minute(s) etc.).
                - positional arguments: arg0, arg1, arg2 etc. Will name
                    positional arguments (if any) of function that uses
                    this decorator. If function has __self__ attribute,
                    it will be used for arg0.
                - keyword arguments: keyword arguments (if any) of
                    function that uses this decorator.
                (default: {"'{f_name}' took {dur:.2f} {uom_name}"})
            uom: time unit of measure to output. Possible values:
                'ms' (milliseconds), 's' (seconds), 'm' (minutes), 'h'
                (hours) (default: {'s'})
        """
        self.log_level = log_level
        self.msg_fmt = msg_fmt
        self.uom = uom

    def _get_message_vals(self, func, duration, args=None, kwargs=None):

        def insert_args(args, func, vals):
            pos = 0
            if hasattr(func, '__self__'):
                vals['arg0'] = func.__self__
                # Move position to right, because arg0 is now already
                # assigned.
                pos = 1
            for i, arg in enumerate(args, pos):
                vals['arg%s' % i] = arg

        if not args:
            args = ()
        if not kwargs:
            kwargs = {}
        uom_name, uom_ratio = self._time_map[self.uom]
        # Define main vals.
        vals = {
            '_f_name': func.__name__,
            '_dur': duration * uom_ratio,
            '_uom_name': uom_name
        }
        insert_args(args, func, vals)
        # Insert kwargs. Will overwrite old keys if some already exist.
        vals.update(kwargs)
        return vals

    def _get_message(
        self,
        func,
        duration: float,
        args: Optional[tuple] = None,
            kwargs: Optional[dict] = None):
        vals = self._get_message_vals(func, duration, args=args, kwargs=kwargs)
        return self.msg_fmt.format(**vals)

    def output(self, func, duration, args=None, kwargs=None):
        msg = self._get_message(func, duration, args=args, kwargs=kwargs)
        if self.log_level:
            log_level = self.log_level.lower()
            getattr(logging, log_level)(msg)
        else:
            print(msg)

    def __call__(self, func):
        """Wrap function and track time it took to execute it."""
        def wrapped(*args, **kwargs):
            ts = time.time()
            res = func(*args, **kwargs)
            te = time.time()
            self.output(func, te-ts, args=args, kwargs=kwargs)
            return res
        return wrapped


class _CatchExceptions(object):
    """Decorator class to catch exceptions on function run."""

    def __init__(self, exceptions_lst, err_msg_pos=0):
        """Init exceptions config."""
        self.exceptions_lst = exceptions_lst
        self.err_msg_pos = err_msg_pos

    def _get_exceptions(self):
        # tuple is required when using exceptions in catching.
        return tuple(
            [exc_item['exception'] for exc_item in self.exceptions_lst])

    @staticmethod
    def _is_bound_method(method):
        return hasattr(method, '__self__')

    @classmethod
    def _map_args_kwargs(cls, func, *fn_args, **fn_kwargs):
        def combine_args_kwargs(args, kwargs):
            return {'args': tuple(args), 'kwargs': kwargs}

        # Make copies from args and kwargs to not modify original
        # arguments passed. Shallow copy is enough here, because values
        # itself are not modified, only add and remove operations are
        # made.
        args = list(fn_args)  # converted back to tuple in the end.
        kwargs = dict(fn_kwargs)
        spec = inspect.getfullargspec(func)
        # If method is bound, it means first argument is an instance of
        # some class. In that case we insert it into args as first
        # argument.
        if cls._is_bound_method(func):
            args.insert(0, func.__self__)
        # If we have *args, it means kwargs can't be specified as
        # positional arguments or if we do not have defaults, it means
        # only positional arguments can be used as positional. In both
        # cases, no further mapping is needed.
        if spec.varargs or not spec.defaults:
            return combine_args_kwargs(args, kwargs)
        # Get number of kw and real positional arguments.
        kwargs_num = len(spec.defaults)
        args_num = len(spec.args)
        real_args_num = args_num - kwargs_num
        # Check if real positional arguments match all positional
        # arguments. If it does, it means all kwargs were used as kwargs
        # and not as positional arguments.
        fake_args_num = len(args) - real_args_num
        if not fake_args_num:
            return combine_args_kwargs(args, kwargs)
        # Find kwargs that are used as positional arguments.
        args = list(args)
        # pop fake positional arguments and put it back into kwargs.
        # Need to start from the end of index range, because poping
        # items in increasing order, would mess up index.
        for index in range(real_args_num+fake_args_num-1, real_args_num-1, -1):
            kwargs[spec.args[index]] = args.pop(index)
        return combine_args_kwargs(args, kwargs)

    def __call__(self, fn, *args, **kwargs):
        """Catch defined exceptions and output specified message."""
        def new_func(*args, **kwargs):
            # TODO: allow specifying __self__ attributes to be used
            # in message.
            def parse_params(e, params=False):
                args_kwargs = self._map_args_kwargs(fn, *args, **kwargs)
                if not params:
                    params = []
                values = []
                # key for args is index, for kwargs, it is dictionary
                # key.
                for key, _type in params:
                    # Access value from args or kwargs and add it to
                    # list.
                    values.append(args_kwargs[_type][key])
                # Combine with error message if error_msg_pos is not
                # False.
                if self.err_msg_pos is not False:
                    values.insert(self.err_msg_pos, e)
                return tuple(values)

            exceptions = self._get_exceptions()
            try:
                return fn(*args, **kwargs)
            except exceptions as e:
                for exc_dct in self.exceptions_lst:
                    exception = exc_dct['exception']
                    # Check if specific condition matches what is defined
                    # in exception catching config.
                    if isinstance(e, exception):
                        exc_condition = exc_dct.get('exc_condition')
                        # Check if exception has condition (such as specific
                        # error code) and run it.
                        if not exc_condition or exc_condition(e):
                            exc_to_raise = (
                                exc_dct.get('raise_exception') or exception)
                            values = parse_params(e, exc_dct.get('params'))
                            raise exc_to_raise(exc_dct['msg'] % values)
                raise
        return new_func


time_it = _TimeIt
catch_exceptions = _CatchExceptions
